from .base_datatype_mapper_test import *
from .string_mapper_test import *
from .set_mapper_decorator_test import *
from .model_mapper_test import *
from .dict_mapper_test import *
from .list_mapper_test import *
from .uniform_list_mapper_test import *
from .byte_mapper_test import *



from mock import Mock
from tests.helpers.testbase import TestBase
from cerami.response import SaveResponse

class TestSaveResponse(TestBase):
    def test__init__(self):
        """it calls reconstructor with the raw Attributes"""
        mocked_reconstructor = Mock()
        mocked_reconstructor.reconstruct.return_value = 'mocked'
        db_response = {'Attributes': {'test': True}}
        resp = SaveResponse(db_response, mocked_reconstructor)
        assert resp.item == 'mocked'
        mocked_reconstructor.reconstruct.assert_called_with(
            db_response['Attributes'])

    def test__init__keyerror(self):
        """it sets item to know when missing Attributes"""
        mocked_reconstructor = Mock()
        db_response = {}
        resp = SaveResponse(db_response, mocked_reconstructor)
        assert resp.item == None





from .model import Model

class Cerami(object):
    def __init__(self, client):
        self.client = client

    @property
    def Model(self):
        """return a model with a connected client"""
        model_cls = Model
        model_cls.client = self.client
        return model_cls



from .response import Response

class SearchResponse(Response):
    def __init__(self, response, reconstructor):
        super(SearchResponse, self).__init__(response, reconstructor)
        self.count = self._raw['Count']
        self.scanned_count = self._raw['ScannedCount']
        self.last_evaluated_key = self._raw.get('LastEvaluatedKey')
        self._items = self._raw.get('Items', [])

    @property
    def items(self):
        for item in self._items:
            yield self.reconstructor.reconstruct(item)



import dateutil.parser
import bson
import numbers
from copy import copy
from datetime import datetime
from bson.objectid import InvalidId
from .mapper import (
    StringMapper,
    ModelMapper,
    DictMapper,
    ListMapper,
    UniformListMapper,
    ByteMapper,
    SetMapperDecorator)
from .expression import (
    BaseExpression,
    ArithmeticExpression,
    ListAppendExpression,
    EqualityExpression)

class DefaultMapGuesser(object):
    def guess(self, key, value):
        """guess the datatype from the value
        Guessing for mapping means we are trying to figure out the datatype
        in order to convert it into a dict usable by dynamodb. So this
        guesser will makes its guess based on value directly
        """
        if isinstance(value, numbers.Number):
            return Number()
        elif isinstance(value, dict):
            return Map()
        elif isinstance(value, list):
            return List()
        else:
            return String()

class DefaultParseGuesser(object):
    def guess(self, key, value):
        """guess the datatype from the key within value
        we are guessing on something like {'M': {'test': {'S': 'hello'}}}
        where key is 'test' and value is {'S': 'hello'}
        this will fetch the inner key ('S') and guess from this value
        """
        attr_key = list(value)[0]
        if attr_key == "N":
            return Number()
        elif attr_key == "S":
            return String()
        elif attr_key == "SS":
            return Set(String())
        elif attr_key == "NS":
            return Set(Number())
        elif attr_key == "L":
            return List()
        elif attr_key == "M":
            return Map()
        else:
            import pdb; pdb.set_trace()
            return String()

class DynamoDataType(object):
    """Abstract class for all DataTypes
    class variable defined. It is used for querying to
    determine what type of attribute is being searched upon
    """
    def __init__(self, default=None, column_name=""):
        self.default = default
        self.set_column_name(column_name)

    def eq(self, value):
        return EqualityExpression("=", self, value)

    def neq(self, value):
        return EqualityExpression("!=", self, value)

    def gt(self, value):
        return EqualityExpression(">", self, value)

    def gte(self, value):
        return EqualityExpression(">=", self, value)

    def lt(self, value):
        return EqualityExpression("<", self, value)

    def lte(self, value):
        return EqualityExpression("<=", self, value)

    def set_column_name(self, val):
        self.column_name = val

    def _get_default(self, val=None):
        if self.default:
            if callable(self.default):
                return self.default()
            else:
                return self.default
        return None


class BaseString(DynamoDataType):
    @property
    def condition_type(self):
        return "S"

    @property
    def mapper(self):
        return StringMapper(self)


class BaseNumber(DynamoDataType):
    @property
    def condition_type(self):
        return "N"

    def add(self, value):
        return ArithmeticExpression('+', self, value)

    def subtract(self, value):
        return ArithmeticExpression('-', self, value)

    @property
    def mapper(self):
        return StringMapper(self)

class ByteBuffer(DynamoDataType):
    @property
    def condition_type(self):
        return "B"

    @property
    def mapper(self):
        return ByteMapper(self)

    def build(self, val):
        val = val or self._get_default()
        if isinstance(val, bytes) or val is None:
            return val
        return val.encode('utf-8')

    def as_item(self, val):
        return val

    def as_dict(self, val):
        return self.as_item(val)


class Number(BaseNumber):
    def build(self, val):
        val = val or self._get_default()
        return val

    def as_item(self, val):
        return val

    def as_dict(self, val):
        return self.as_item(val)

class String(BaseString):
    def build(self, val):
        val = val or self._get_default()
        return val

    def as_item(self, val):
        return str(val)

    def as_dict(self, val):
        return self.as_item(val)


class Datetime(BaseString):
    def build(self, val):
        val = val or self._get_default()
        if isinstance(val, datetime) or val is None:
            return val
        return dateutil.parser.parse(val)

    def as_item(self, val):
        return val.isoformat()

    def as_dict(self, val):
        return self.as_item(val)

class Map(DynamoDataType):
    def __init__(self, map_guesser=None, parse_guesser=None, default=None, **kwargs):
        super(Map, self).__init__(**kwargs)
        self.map_guesser = map_guesser or DefaultMapGuesser()
        self.parse_guesser = parse_guesser or DefaultParseGuesser()

    @property
    def condition_type(self):
        return "M"

    @property
    def mapper(self):
        return DictMapper(self, self.map_guesser, self.parse_guesser)

    def key(self, datatype, key):
        column_name = self.column_name + "." + key
        return type(datatype)(column_name=column_name)

    def build(self, val):
        val = val or self._get_default()
        if val == None or isinstance(val, dict):
            return val
        else:
            raise ValueError("build must receive a dict")

    def as_dict(self, val):
        if val == None or isinstance(val, dict):
            return val
        else:
            raise ValueError("as_dict must receive a dict")

    def as_item(self, val):
        if val == None or isinstance(val, dict):
            return self.mapper.map(val)
        else:
            raise ValueError("as_item must receive a dict")

class ModelMap(DynamoDataType):
    def __init__(self, model_cls, default=None, **kwargs):
        super(ModelMap, self).__init__(**kwargs)
        self.model_cls = model_cls
        for column in self.model_cls._columns:
            setattr(self, column.column_name, copy(column))

    @property
    def condition_type(self):
        return "M"

    @property
    def mapper(self):
        return ModelMapper(self)

    def build(self, val):
        val = val or self._get_default()
        if val == None:
            return None
        if isinstance(val, dict):
            return self.model_cls(data=val)
        elif isinstance(val, self.model_cls):
            return val
        else:
            raise ValueError("build must receive a dict or Model")

    def as_dict(self, val):
        return val.as_dict()

    def as_item(self, val):
        return val.as_item()

    def set_column_name(self, val):
        super(ModelMap, self).set_column_name(val)
        for name, attr in self.__dict__.items():
            if isinstance(attr, DynamoDataType):
                new_name = val + "." + name
                attr.set_column_name(new_name)


class List(DynamoDataType):
    def __init__(self, map_guesser=None, parse_guesser=None, default=None, **kwargs):
        super(List, self).__init__(**kwargs)
        self.map_guesser = map_guesser or DefaultMapGuesser()
        self.parse_guesser = parse_guesser or DefaultParseGuesser()

    @property
    def condition_type(self):
        return "L"

    @property
    def mapper(self):
        return ListMapper(self, self.map_guesser, self.parse_guesser)

    def append(self, array):
        if not isinstance(array, list):
            array = [array]
        return ListAppendExpression(self, array)

    def index(self, idx, datatype_cls):
        """return an expression attribute of the inner datatype
        sets the index value on the expression attribute 
        """
        dt = datatype_cls(column_name=self.column_name)
        dt._index = idx
        return dt

    def build(self, val):
        val = val or self._get_default()
        if val == None or isinstance(val, list):
            return val
        else:
            raise ValueError("build must receive a list")

    def as_dict(self, val):
        if isinstance(val, list):
            return val
        else:
            raise ValueError("as_dict must receive a list")

    def as_item(self, val):
        if isinstance(val, list):
            return self.mapper.map(val)
        else:
            raise ValueError("as_item must receive a list")

class UniformList(List):
    def __init__(self, datatype, **kwargs):
        super(UniformList, self).__init__(**kwargs)
        self.datatype = datatype

    @property
    def mapper(self):
        # TODO - create this class
        return UniformListMapper(self.datatype.mapper)

    def build(self, val=None):
        if val == None:
            return val
        if not isinstance(val, list):
            raise ValueError("{} is not a list".format(val))
        return [self.datatype.build(i) for i in val]

    def index(self, idx):
        """return an expression attribute of the inner datatype
        sets the index value on the expression attribute 
        """
        datatype_cls = type(self.datatype)
        dt = datatype_cls(column_name=self.column_name)
        dt._index = idx
        return dt

    def as_item(self, val):
        return [self.datatype.as_item(i) for i in val]

    def as_dict(self, val):
        return [self.datatype.as_dict(i) for i in val]

class Set(DynamoDataType):
    def __init__(self, datatype, default=None, **kwargs):
        super(Set, self).__init__(**kwargs)
        self.datatype = datatype

    @property
    def condition_type(self):
        return self.datatype.condition_type + "S"

    @property
    def mapper(self):
        return SetMapperDecorator(self.datatype.mapper)

    def build(self, val):
        if val == None:
            return None
        if not isinstance(val, list):
            raise ValueError("{} is not a list".format(val))
        return [self.datatype.build(i) for i in val]

    def as_item(self, val):
        """another thing i dont like
        i am essentially recreating the entire list when
        as_item is called, so if it is a long list of strings
        this is a complete waste of time. Pretty stupid...

        but it is really powerful when nested objects come into
        play. so probably the simplest thing would be to have
        some way to check if the datatype is a literal like a 
        string or number and if so just return the value
        """
        return [self.datatype.as_item(i) for i in val]

    def as_dict(self, val):
        return self.as_item(val)



class BaseDatatypeMapper(object):
    def __init__(self, datatype):
        self.datatype = datatype

    def map(self, value):
        """return the value and its condition_type
        ---
        ex) {"S": "Zac"}

        """
        res = {}
        res[self.datatype.condition_type] = self.resolve(value)
        return res

    def resolve(value):
        """returns the value resolved for dynamodb
        that is to say it is how the database expects
        it to be passed for all of the operations
        """
        pass

    def parse(self, mapped_dict):
        """extract the value from the mapped_dict
        This is the opposite of mapping. So when given
        {"S": "Zac"} it will return "Zac"
        """
        if mapped_dict.get('NULL') == True:
            return None
        return mapped_dict[self.datatype.condition_type]



from mock import Mock
from tests.helpers.testbase import TestBase
from cerami.datatype.mapper import UniformListMapper


class TestUniformListMapper(TestBase):
    def setUp(self):
        self.mocked_mapper = Mock()
        self.mapper = UniformListMapper(self.mocked_mapper)

    def test__init__(self):
        """it sets mapper"""
        assert self.mapper.mapper == self.mocked_mapper

    def test_map(self):
        """it calls mocked_map for each item in the list and returns the dict"""
        self.mocked_mapper.map.return_value = 'mocked'
        val = [1]
        assert self.mapper.map(val) == {'L': ['mocked']}
        self.mocked_mapper.map.assert_called_with(1)

    def test_resolve(self):
        """it calls mocked_resolve for each value"""
        self.mocked_mapper.resolve.return_value = 'mocked'
        val = [1]
        assert self.mapper.resolve(val) == ['mocked']
        self.mocked_mapper.resolve.assert_called_with(1)

    def test_parse(self):
        """it calls mapper.parse for each item in the list"""
        self.mocked_mapper.parse.return_value = 'mocked'
        mapped_dict = {'L': [{'S': 'test'}]}
        assert self.mapper.parse(mapped_dict) == ['mocked']
        self.mocked_mapper.parse.assert_called_with({'S': 'test'})



from .search_request import SearchRequest
from .search_attribute import DictAttribute
from ..response import GetResponse


class GetRequest(SearchRequest):
    def execute(self):
        """TODO: need to validate all primary key / saerch keys are present"""
        response = self.client.get_item(**self.build())
        return GetResponse(response, self.reconstructor)

    def key(self, *expressions):
        for expression in expressions:
            key_dict = {}
            key_dict[expression.datatype.column_name] = expression.attribute_map()
            self.add_attribute(DictAttribute, 'Key', key_dict)
        return self



from .base_reconstructor import BaseReconstructor

class ModelReconstructor(BaseReconstructor):
    def __init__(self, model_cls):
        self.model_cls = model_cls

    def reconstruct(self, item_dict):
        data = {}
        for column in self.model_cls._columns:
            try:
                val = item_dict[column.column_name]
            except KeyError:
                continue
            data[column.column_name] = column.mapper.parse(val)
        return self.model_cls(**data)



from .base_datatype_mapper import BaseDatatypeMapper

class UniformListMapper(object):
    def __init__(self, mapper):
        self.mapper = mapper

    def map(self, value):
        mapped_items = [self.mapper.map(i) for i in value]
        return {'L': mapped_items}

    def resolve(self, value):
        return [self.mapper.resolve(i) for i in value]

    def parse(self, mapped_dict):
        condition_type  = self.mapper.datatype.condition_type
        return [self.mapper.parse(i) for i in mapped_dict["L"]]



class UpdateExpressionDecorator(object):
    def __init__(self, action, expression):
        self.action = action
        self.expression = expression

    def __str__(self):
        return "{} {}".format(self.action, self.expression)

    def attribute_map(self):
        return self.expression.attribute_map()

    def value_dict(self):
        return self.expression.value_dict()

    def _generate_variable_name(self, column_name):
        return self.expression._generate_variable_name(self, column_name)





from .search_request import SearchRequest
from ..response import SearchResponse

class ScanRequest(SearchRequest):
    def execute(self):
        response = self.client.scan(**self.build())
        return SearchResponse(response, self.reconstructor)



from mock import patch, call, Mock
from tests.helpers.testbase import TestBase
import cerami.datatype as dt
from cerami.model import Model
from cerami.datatype import String
from cerami.datatype.expression import EqualityExpression
from cerami.request.search_attribute import (
    SearchAttribute,
    DictAttribute,
    QueryExpressionAttribute,)
from cerami.request import SearchRequest

class TestModel(Model):
    test = String()

class TestSearchRequest(TestBase):
    def setUp(self):
        self.client = Mock()
        self.request = SearchRequest(self.client, tablename='testtable')

    def test_add_attribute(self):
        """it calls add for the search attribute"""
        with patch("cerami.request.search_attribute.SearchAttribute.add") as add:
            self.request.add_attribute(SearchAttribute, 'Test', 123)
            assert isinstance(
                self.request.search_attributes['Test'],
                SearchAttribute)
            add.assert_called_with(123)

    def test_filter(self):
        """it adds FilterExpression, ExpresionAttributeNames,
        ExpressionAttributeValues"""
        with patch("cerami.request.SearchRequest.add_attribute") as add:
            expression = TestModel.test.eq('123')
            names = {}
            names[expression.expression_attribute_name] = 'test'

            self.request.filter(expression)
            calls = [
                call(QueryExpressionAttribute,
                     'FilterExpression',
                     expression),
                call(DictAttribute,
                     'ExpressionAttributeNames',
                     names),
                call(DictAttribute,
                     'ExpressionAttributeValues',
                     expression.value_dict())
            ]
            add.assert_has_calls(calls)

    def test_build(self):
        """returns a dict for all search_attributes"""
        self.request.add_attribute(SearchAttribute, 'Test', 123)
        built = self.request.build()
        expected = {
            "TableName": "testtable",
            "Test": 123,
        }
        assert built == expected

    def test_return_values(self):
        """it calls add_attribute with the value"""
        self.request.add_attribute = Mock()
        self.request.return_values('test')
        self.request.add_attribute.assert_called_with(
            SearchAttribute,
            'ReturnValues',
            'test')



from .base_datatype_mapper import BaseDatatypeMapper
from .string_mapper import StringMapper
from .set_mapper_decorator import SetMapperDecorator
from .model_mapper import ModelMapper
from .dict_mapper import DictMapper
from .list_mapper import ListMapper
from .uniform_list_mapper import UniformListMapper
from .byte_mapper import ByteMapper



from tests.helpers.testbase import TestBase
from cerami.request.search_attribute import SearchAttribute

class TestSearchAttribute(TestBase):
    def setUp(self):
        self.attribute = SearchAttribute("Test")

    def test_add(self):
        """it sets the value"""
        self.attribute.add(123)
        assert self.attribute.value == 123

    def test_build(self):
        """it returns the value"""
        self.attribute.add(123)
        assert self.attribute.build() == 123




import cerami.datatype as dt
from mock import patch, Mock
from tests.helpers.testbase import TestBase
from cerami.reconstructor import ModelReconstructor
from cerami.data_attribute import DynamoDataAttribute
from cerami.request import (
    GetRequest,
    UpdateRequest,
    QueryRequest,
    ScanRequest)
from cerami import Dynamo
from cerami.model import Model


class TestModelClass(TestBase):
    class DummyModel(Model):
        __tablename__ = "test"
        client = Mock()
        _id = dt.String()
        name = dt.String()
        tags = dt.Set(dt.String())

    def setUp(self):
        super(TestModelClass, self).setUp()
        self.model = self.DummyModel(_id='1', name='test')

    def test__new__(self):
        """All Model columns should have a column_name"""
        assert self.DummyModel._id.column_name == "_id"
        assert self.DummyModel.name.column_name == "name"

    def test__column(self):
        """it chould only return the DataTypes on the Schema"""
        idColumn = dt.String()
        testColumn = dt.String()
        class TestModel(Model):
            _id = idColumn
            test = testColumn
        assert len(TestModel._columns) == 2
        assert idColumn in TestModel._columns
        assert testColumn in TestModel._columns


    def test_put_calls_put(self):
        """put should call PutInterface.execute with the data as an item"""
        with patch('cerami.request.put_request.PutRequest.execute') as execute,\
            patch('cerami.request.put_request.PutRequest.item') as item:
            self.model.put()
            item.assert_called_with(self.model.as_item())
            execute.assert_called()

    def test_has_scan_property(self):
        """the class should have a ScanRequest"""
        assert isinstance(self.DummyModel.scan, ScanRequest)
        assert isinstance(self.DummyModel.scan.reconstructor, ModelReconstructor)

    def test_has_query_property(self):
        """the class should have a QueryRequest"""
        assert isinstance(self.DummyModel.query, QueryRequest)
        assert isinstance(self.DummyModel.query.reconstructor, ModelReconstructor)

    def test_has_update_property(self):
        """the class should have an UpdateRequest"""
        assert isinstance(self.DummyModel.update, UpdateRequest)
        assert isinstance(self.DummyModel.update.reconstructor, ModelReconstructor)

    def test_has_get_property(self):
        """the class should have a GetRequest"""
        assert isinstance(self.DummyModel.get, GetRequest)
        assert isinstance(self.DummyModel.get.reconstructor, ModelReconstructor)

    def test_init_sets_attributes(self):
        """it sets all keys in the model"""
        for column in self.model._columns:
            assert hasattr(self.model, column.column_name)


    def test__getattribute__(self):
        """it should return the value passed normally"""
        model = self.DummyModel(_id='abc')
        assert model._id == 'abc'

    def test__getattribute__full(self):
        """it should return the DynamoDataAttribute when full is true"""
        res = self.model.__getattribute__('_id', full=True)
        assert isinstance(res, DynamoDataAttribute)

    def test__setattr__(self):
        """it should call DynamoDataAttribute.set"""
        with patch("cerami.data_attribute.DynamoDataAttribute.set") as setter:
            self.model._id = "new_id"
            setter.assert_called()

    def test_as_item_serializes_each_model_key(self):
        """it should return the data as a dict

        this means called serialize() for each attribute
        """
        with patch("cerami.data_attribute.DynamoDataAttribute.as_dict") as as_dict:
            item = self.model.as_item()
            as_dict.call_count == 3

    def test_get_full_attribute_none(self):
        """it should return None when the key is missing"""
        assert self.model._get_full_attribute('bad') is None



from mock import Mock, patch
from tests.helpers.testbase import TestBase
from cerami.response import SearchResponse
from cerami.request import QueryRequest

class TestQueryRequest(TestBase):
    def setUp(self):
        self.mocked_client = Mock()
        self.request  = QueryRequest(
            tablename="test",
            client=self.mocked_client)

    def test_execute(self):
        """it calls query with the build
        it returns a SearchResponse"""
        with patch("cerami.request.search_request.SearchRequest.build") as build:
            expected = {"fake": True}
            self.mocked_client.query.return_value = {
                'Count': 0,
                'ScannedCount': 0}
            build.return_value = expected
            res = self.request.execute()
            self.mocked_client.query.assert_called_with(fake=True)
            assert isinstance(res, SearchResponse)




from .search_request import SearchRequest
from ..response import SearchResponse
from .search_attribute import (
    SearchAttribute,
    DictAttribute,
    QueryExpressionAttribute)

class QueryRequest(SearchRequest):
    def execute(self):
        response = self.client.query(**self.build())
        return SearchResponse(response, self.reconstructor)

    def key(self, *expressions):
        """return a new SearchInterface setup with query attributes
        KeyConditionExpression, ExpressionAttributeNames,
        ExpressionAttributeValues are all required to query properly

        *expressions are a list of BaseExpressions
        """
        for expression in expressions:
            names = {}
            names[expression.expression_attribute_name] = expression.datatype.column_name
            self.add_attribute(QueryExpressionAttribute,
                              'KeyConditionExpression',
                              expression)
            self.add_attribute(DictAttribute,
                              'ExpressionAttributeNames',
                              names)
            self.add_attribute(DictAttribute,
                              'ExpressionAttributeValues',
                              expression.value_dict())
        return self



from mock import Mock, patch
from datetime import datetime
from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype.expression import (
    EqualityExpression,
    ListAppendExpression,
    ArithmeticExpression)
from cerami.datatype.mapper import (
    ByteMapper,
    StringMapper,
    ModelMapper,
    DictMapper,
    ListMapper,
    SetMapperDecorator)
from cerami.datatype import (
    DynamoDataType,
    BaseString,
    BaseNumber,
    ByteBuffer,
    String,
    Datetime,
    Map,
    ModelMap,
    List,
    Set)

class TestBaseString(TestBase):
    def setUp(self):
        super(TestBaseString, self).setUp()
        self.dt = BaseString(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "S"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, StringMapper)

class TestBaseNumber(TestBase):
    def setUp(self):
        super(TestBaseNumber, self).setUp()
        self.dt = BaseNumber(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "N"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, StringMapper)

    def test_add(self):
        """it returns an ArithmeticExpression"""
        res = self.dt.add(5)
        assert isinstance(res, ArithmeticExpression)
        assert res.expression == '+'

    def test_subtract(self):
        """it returns an ArithmeticExpression"""
        res = self.dt.subtract(5)
        assert isinstance(res, ArithmeticExpression)
        assert res.expression == '-'

class TestString(TestBase):
    def setUp(self):
        super(TestString, self).setUp()
        self.dt = String(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "S"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, StringMapper)

    def test_as_item(self):
        """it returns the value as a string"""
        assert self.dt.as_item(1) == "1"

    def test_build_returns_val(self):
        """it returns the value when it is present"""
        assert self.dt.build("test") == "test"

    def test_build_calls_get_default(self):
        """it calls _get_default when value is falsey"""
        self.dt._get_default = Mock()
        self.dt.build(None)
        self.dt._get_default.assert_called()

class TestByteBuffer(TestBase):
    def setUp(self):
        super(TestByteBuffer, self).setUp()
        self.dt = ByteBuffer(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "B"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, ByteMapper)

    def test_as_item(self):
        assert self.dt.as_item(b'1') == b'1'

    def test_as_dict(self):
        assert self.dt.as_dict(b'1') == b'1'

    def test_build_returns_val(self):
        """it returns the value when it is a byte"""
        val = b'hello'
        assert self.dt.build(val) == val

    def test_build_encodes(self):
        """it encodes the value when it is not a byte"""
        val = 'hello'
        expected = val.encode('utf-8')
        assert self.dt.build(val) == expected

class TestDatetime(TestBase):
    def setUp(self):
        super(TestDatetime, self).setUp()
        self.dt = Datetime(column_name="test")

    def test_condition_type(self):
        assert self.dt.condition_type == "S"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, StringMapper)

    def test_as_item(self):
        """it returns the value in isoformat"""
        val = Mock()
        val.isoformat.return_value = "fake isostring"
        res = self.dt.as_item(val)
        val.isoformat.assert_called()
        assert res == "fake isostring"

    def test_build_val_datetime(self):
        """it returns the passed value when it is a datetime"""
        now = datetime.now()
        assert self.dt.build(now) == now

    def test_build_calls_get_default(self):
        """it calls _get_default when value is falsey"""
        self.dt._get_default = Mock(return_value=None)
        self.dt.build(None)
        self.dt._get_default.assert_called()

    def test_build_parses_datetime_string(self):
        """it parses the datetime string"""
        now_str = datetime.now().isoformat()
        assert isinstance(self.dt.build(now_str), datetime)

class TestMap(TestBase):
    def setUp(self):
        super(TestMap, self).setUp()
        self.mocked_guesser = Mock()
        self.dt = Map(
            map_guesser=self.mocked_guesser,
            parse_guesser=self.mocked_guesser)

    def test__init__(self):
        """it initializes the guesser"""
        assert self.dt.map_guesser == self.mocked_guesser
        assert self.dt.parse_guesser == self.mocked_guesser

    def test_condition_type(self):
        assert self.dt.condition_type == "M"

    def test_mapper(self):
        """it is an instance of DictMapper"""
        assert isinstance(self.dt.mapper, DictMapper)

    def test_build(self):
        """it returns the val when it is a dict"""
        val = {'test': 1}
        assert self.dt.build(val) == val

    def test_as_dict(self):
        """it returns the val when it is a dict"""
        val = {'test': 1}
        assert self.dt.as_dict(val) == val

    def test_as_item(self):
        """it returns mapper.map's value"""
        with patch('cerami.datatype.mapper.DictMapper.map') as mp:
            mp.return_value = 'mocked'
            val = {'test': 1}
            assert self.dt.as_item(val) == 'mocked'
            mp.assert_called_with(val)

class TestList(TestBase):
    def setUp(self):
        super(TestList, self).setUp()
        self.mocked_guesser = Mock()
        self.dt = List(
            column_name="test",
            map_guesser=self.mocked_guesser,
            parse_guesser=self.mocked_guesser)

    def test__init__(self):
        """it sets guesser"""
        assert self.dt.map_guesser == self.mocked_guesser
        assert self.dt.parse_guesser == self.mocked_guesser

    def test_property(self):
        assert self.dt.condition_type == "L"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, ListMapper)

    def test_append(self):
        """it should return a ListAppendExpression"""
        res = self.dt.append(['Test'])
        assert isinstance(res, ListAppendExpression)
        assert res.value == ['Test']

    def test_append_scalar(self):
        """it converts the value to a list when it isnt one"""
        res = self.dt.append('Test')
        assert isinstance(res, ListAppendExpression)
        assert res.value == ['Test']

    def test_index(self):
        """it returns a new instance of the datatype passed with the column_name
        it sets _index on the returned datatype instance"""
        res = self.dt.index(0, String)
        assert isinstance(res, String)
        assert res._index == 0

    def test_build(self):
        """it returns the val passed when it is a list"""
        val = [1,2,3]
        assert self.dt.build(val) == val

    def test_as_dict(self):
        """it returns the val passed when it is a list"""
        val = [1,2,3]
        assert self.dt.build(val) == val

    def test_as_item(self):
        """it reutns the mapped value when it is a list"""
        with patch('cerami.datatype.mapper.ListMapper.map') as lm:
            lm.return_value = 'mocked'
            val = [1,2,3]
            assert self.dt.as_item(val) == 'mocked'

class TestModelMap(TestBase):
    class TestModel(Model):
        __tablename__ = "test"
        test1 = String()
        test2 = String()

    def setUp(self):
        super(TestModelMap, self).setUp()
        self.dt = ModelMap(self.TestModel, column_name="test")

    def test__init__(self):
        """it should setattr for each column in the model_cls"""
        for column in self.TestModel._columns:
            assert hasattr(self.dt, column.column_name)
            assert isinstance(getattr(self.dt, column.column_name),
                              type(column))

    def test_condition_type(self):
        assert self.dt.condition_type == "M"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, ModelMapper)

    def test_build_dict(self):
        """it returns a model_cls instance when val is a dict"""
        data = {'test1': 'test', 'test2': 1}
        assert isinstance(self.dt.build(data), self.TestModel)

    def test_build_model(self):
        """it returns the value when it is already a model_cls instance"""
        model = self.TestModel()
        assert self.dt.build(model) == model

    def test_as_dict(self):
        """it calls the passed value as_dict method"""
        val = Mock()
        self.dt.as_dict(val)
        val.as_dict.assert_called()

    def test_as_item(self):
        """it calls the passed value as_item method"""
        val = Mock()
        self.dt.as_item(val)
        val.as_item.assert_called()

    def set_column_name(self):
        """it updates the column_name and all nested column names"""
        new_name = "test2"
        self.dt.set_column_name(new_name)
        assert self.dt.column_name == new_name
        for column in TestModel._columns:
            nested_column = getattr(self.dt, column.column_name)
            new_nested_name = new_name + "." + column.column_name
            assert nested_column.column_name == new_nested_name

class TestSet(TestBase):
    def setUp(self):
        super(TestSet, self).setUp()
        self.dt = Set(String(), column_name="test")

    def test_has_a_datatype(self):
        assert isinstance(self.dt.datatype, String)

    def test_condition_type(self):
        """it adds an S to the datatype's condition_type"""
        assert self.dt.condition_type == "SS"

    def test_mapper(self):
        assert isinstance(self.dt.mapper, SetMapperDecorator)

class TestDynamoDataType(TestBase):
    def setUp(self):
        super(TestDynamoDataType, self).setUp()
        self.dt = DynamoDataType(column_name="test")

    def test_eq(self):
        res = self.dt.eq(1)
        assert isinstance(res, EqualityExpression)
        assert res.expression == "="
        
    def test_neq(self):
        res = self.dt.neq(1)
        assert isinstance(res, EqualityExpression)
        assert res.expression == "!="

    def test_gt(self):
        res = self.dt.gt(1)
        assert isinstance(res, EqualityExpression)
        assert res.expression == ">"

    def test_gte(self):
        res = self.dt.gte(1)
        assert isinstance(res, EqualityExpression)
        assert res.expression == ">="

    def test_lt(self):
        res = self.dt.lt(1)
        assert isinstance(res, EqualityExpression)
        assert res.expression == "<"

    def test_lte(self):
        res = self.dt.lte(1)
        assert isinstance(res, EqualityExpression)
        assert res.expression == "<="



from tests.helpers.testbase import TestBase
from cerami.datatype import ByteBuffer
from cerami.datatype.mapper import ByteMapper

class TestByteMapper(TestBase):
    def setUp(self):
        self.dt = ByteBuffer()
        self.mapper = ByteMapper(self.dt)

    def test_resolve(self):
        """it returns the value as a string"""
        assert self.mapper.resolve('hello') == 'hello'.encode('utf-8')




from .search_attribute import SearchAttribute

class RemoveExpressionAttribute(SearchAttribute):
    def __init__(self, name, value=None):
        value = value or []
        super(RemoveAttribute, self).__init__(name, value)

    def add(self, expression):
        self.value.append(expression)

    def build(self):
        return "REMOVE " + ', '.join(str(expr) for expr in self.value)






from .search_attribute import SearchAttribute

class UpdateExpressionAttribute(SearchAttribute):
    def __init__(self, name, value=None):
        value = value or {}
        super(UpdateExpressionAttribute, self).__init__(name, value)

    def add(self, update_action):
        if self.value.get(update_action.action):
            self.value[update_action.action].append(update_action.expression)
        else:
            self.value[update_action.action] = [update_action.expression]

    def build(self):
        """return all grouped expressions"""
        operations = []
        for action, expressions in self.value.items():
            operation = action + ' ' + ', '.join(str(expr) for expr in expressions)
            operations.append(operation)
        return ' '.join(o for o in operations)



from .base_datatype_mapper import BaseDatatypeMapper

class SetMapperDecorator(object):
    def __init__(self, mapper):
        self.mapper = mapper

    def map(self, value):
        res = {}
        res[self.mapper.datatype.condition_type + "S"] = self.resolve(value)
        return res

    def resolve(self, value):
        return [self.mapper.resolve(i) for i in value]

    def parse(self, mapped_dict):
        condition_type  = self.mapper.datatype.condition_type + "S"
        return [i for i in mapped_dict[condition_type]]



from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype import String
from cerami.decorators import primary_key

class TestPrimaryKey(TestBase):
    @primary_key('id')
    class SinglePartition(Model):
        id = String()

    @primary_key('id', 'name')
    class SortKeyModel(Model):
        id = String()
        name = String()

    def test_single_key(self):
        cls = self.SinglePartition
        assert cls._primary_key == (cls.id,)

    def test_sort_key(self):
        cls = self.SortKeyModel
        assert cls._primary_key == (cls.id, cls.name)



import random
import string

class BaseExpression(object):
    def __init__(self, expression, datatype, value):
        """
        based on the docs, expression_attribute_names start with a #
        expression_attribute_value names start with a :
        """
        self.expression = expression
        self.datatype = datatype
        self.value = value
        column_name_safe = datatype.column_name.replace('.', '_')
        self.expression_attribute_name = "#__{}".format(column_name_safe)
        self.expression_attribute_value_name = self._generate_variable_name(
            column_name_safe)

    def __str__(self):
        attr_name = self.expression_attribute_name
        if hasattr(self.datatype, '_index'):
            attr_name = "{}[{}]".format(attr_name, self.datatype._index)
        return "{attr_name} {expression} {value_name}".format(
            attr_name=attr_name,
            expression=self.expression,
            value_name=self.expression_attribute_value_name)

    def attribute_map(self):
        """return the value and its condition_type
        by calling a DatatypeMapper
        """
        return self.datatype.mapper.map(self.value)

    def value_dict(self):
        """return the expected dict for expression-attribute-values
        { ":value_name" : { "condition_type": "value" } }
        ---
        ex) {":name":{"S": "Zac"}}
        """
        res = {}
        res[self.expression_attribute_value_name] = self.datatype.mapper.map(self.value)
        return res

    def _generate_variable_name(self, column_name):
        letters = string.ascii_lowercase
        random_letters = ''.join(random.choice(letters) for i in range(5))
        return ":_{}_{}".format(column_name, random_letters)



from .search_attribute_test import TestSearchAttribute
from .dict_attribute_test import TestDictAttribute
from .update_expression_attribute_test import TestUpdateExpressionAttribute
from .query_expression_attribute_test import TestQueryExpressionAttribute



from .scan_request  import ScanRequest
from .put_request import PutRequest
from .update_request import UpdateRequest
from .get_request import GetRequest
from .search_request import SearchRequest
from .query_request import QueryRequest
from .delete_request import DeleteRequest



from mock import patch
from tests.helpers.testbase import TestBase
from cerami.datatype import String
from cerami.datatype.mapper import BaseDatatypeMapper

class TestBaseDatatypeMapper(TestBase):
    def setUp(self):
        self.dt = String()
        self.mapper = BaseDatatypeMapper(self.dt)

    def test_map(self):
        """it returns a dict
        with the key the condition_type
        and the value the result of resolve()
        """
        with patch('cerami.datatype.mapper.BaseDatatypeMapper.resolve') as resolve:
            resolve.return_value = "mocked"
            res = self.mapper.map('test')
            assert res == {"S": "mocked"}

    def test_pasrse(self):
        """it returns the value for the key corresponding to the condition_type"""
        mapped_dict = {'S': 1}
        assert self.mapper.parse(mapped_dict) == 1



from .base_datatype_mapper import BaseDatatypeMapper

class ByteMapper(BaseDatatypeMapper):
    def resolve(self, value):
        """try to encode it or return the value (meaning its already encoded"""
        try:
            return value.encode('utf-8')
        except AttributeError:
            return value




from tests.helpers.testbase import TestBase
from cerami.datatype import String
from cerami.datatype.mapper import StringMapper

class TestStringMapper(TestBase):
    def setUp(self):
        self.dt = String()
        self.mapper = StringMapper(self.dt)

    def test_resolve(self):
        """it returns the value as a string"""
        assert self.mapper.resolve(1) == "1"



from .search_attribute import SearchAttribute
from .dict_attribute import DictAttribute
from .query_expression_attribute import QueryExpressionAttribute
from .update_expression_attribute import UpdateExpressionAttribute
from .update_action import UpdateAction



from .datatype import DynamoDataType
from .data_attribute import DynamoDataAttribute
from .reconstructor import ModelReconstructor
from .request.return_values import ALL_NEW
from .request import (
    GetRequest,
    ScanRequest,
    PutRequest,
    UpdateRequest,
    DeleteRequest,
    QueryRequest)


class ModelMeta(type):
    def __new__(cls, clsname, bases, dct):
        """override the class creation
        * Add the _columns property to store all DataTypes
        * All columns on the Model need to have a string
          reference to the name of the column.
        ---

        For example:
        class MyModel(Model):
            _id: String()
        MyModel._id.column_name ==> "_id"
        """
        dct["_columns"] = []
        for name, val in dct.items():
            if isinstance(val, DynamoDataType):
                val.set_column_name(name.lower())
                dct["_columns"].append(val)
        return super(ModelMeta, cls).__new__(cls, clsname, bases, dct)

    @property
    def get(cls):
        return GetRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

    @property
    def scan(cls):
        return ScanRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

    @property
    def query(cls, *expressions):
        return QueryRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

    @property
    def put(cls):
        return PutRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

    @property
    def update(cls):
        update = UpdateRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))
        return update.return_values(ALL_NEW)


    @property
    def delete(cls):
        return DeleteRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

class Model(object, metaclass=ModelMeta):
    def __init__(self, **data_kwargs):
        """set all column values from data

        It will set any value not present in data but part of
        the models columns to None
        """
        data = data_kwargs or {}
        data_keys = data.keys()
        for column in self._columns:
            name = column.column_name
            value = data.get(name, None)
            data_exists = name in data or column.default
            attr = DynamoDataAttribute(column, value, initialized=data_exists)
            setattr(self, name, attr)

    def __getattribute__(self, key, full=False):
        """override __getattribute__

        most of the time we want to call attr.get()
        when accessing a DynamoDataAttribute. The
        full flag can be used to get the actual object.
        """
        attr = super(Model, self).__getattribute__(key)
        if isinstance(attr, DynamoDataAttribute) and not full:
            return attr.get()
        else:
            return attr

    def __setattr__(self, key, value):
        attr = self._get_full_attribute(key)
        if isinstance(attr, DynamoDataAttribute):
            attr.set(value)
        else:
            super(Model, self).__setattr__(key, value)

    def _get_full_attribute(self, key):
        try:
            return self.__getattribute__(key, full=True)
        except AttributeError:
            return None

    def as_dict(self):
        """return all data values as a dict"""
        item = {}
        for column in self._columns:
            name = column.column_name
            attr = self._get_full_attribute(name)
            if attr.initialized or attr._changed:
                item[name] = attr.as_dict()
        return item

    def as_item(self):
        """return all data values in a format for dynamodb"""
        item = {}
        for column in self._columns:
            name = column.column_name
            attr = self._get_full_attribute(name)
            if attr.initialized or attr._changed:
                item[name] = column.mapper.map(attr.value)
        return item

    def delete(self):
        """delete this record from the database"""
        deleter = self.__class__.delete
        for column in self._primary_key:
            column_name = column.column_name
            deleter = deleter.key(column.eq(getattr(self.data, column_name)))
        return deleter.execute()

    def put(self):
        """add this record to the database"""
        putter = self.__class__.put
        putter.item(self.as_item())
        return putter.execute()

    def update(self):
        """update this record in the database"""
        updater = self.__class__.update
        for column in self._primary_key:
            column_name = column.column_name
            updater.key(column.eq(getattr(self, column_name)))
        for column in self._columns:
            column_name = column.column_name
            attr = self._get_full_attribute(column_name)
            if (not column in self._primary_key
                and (attr.initialized or attr._changed)):
                updater.set(column.eq(attr.as_dict()))
        return updater.execute()



from .base_expression import BaseExpression
from .equality_expression import EqualityExpression
from .arithmetic_expression import ArithmeticExpression
from .list_append_expression import ListAppendExpression
from .update_remove_expression import UpdateRemoveExpression



from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype import (
    String,
    Number,
    ModelMap)
from cerami.datatype.mapper import (
    ModelMapper)

class TestModelMapper(TestBase):
    class TestModel(Model):
        __tablename__ = 'test'
        test1 = String()
        test2 = Number()

    def setUp(self):
        super(TestModelMapper, self).setUp()
        self.dt = ModelMap(self.TestModel)
        self.mapper = ModelMapper(self.dt)

    def test_resolve(self):
        """it should return the a dict with each column its resolved value
        Kinda hard to see..
        But imagine if there was a datetime field in data, it
        should resolve to the string version
        """
        data = {'test1': 'test', 'test2': 2}
        expected = {'test1': 'test', 'test2': '2'}
        res = self.mapper.resolve(data)
        assert res == expected

    def test_map(self):
        """it should return the model as asn item"""
        model = self.TestModel(test1="test", test2=2)
        res = self.mapper.map(model)
        expected = {'M': {'test1': {'S': 'test'}, 'test2': {'N': '2'}}}
        assert res == expected

    def test_parse(self):
        """it returns a model based on the passed dict"""
        mapped_dict = {'M': {'test1': {'S': 'hello'}}}
        res = self.mapper.parse(mapped_dict)
        assert isinstance(res, self.TestModel)
        assert res.test1 == 'hello'
        assert res.test2 == None
        assert res.as_item() == {'test1': {'S': 'hello'}}



from .response import Response

class SaveResponse(Response):
    def __init__(self, response, reconstructor):
        super(SaveResponse, self).__init__(response, reconstructor)
        try:
            self.item = self.reconstructor.reconstruct(self._raw['Attributes'])
        except KeyError:
            self.item = None



class primary_key(object):
    def __init__(self, partition_key, sort_key=None):
        self.partition_key = partition_key
        self.sort_key = sort_key

    def __call__(self, cls):
        column = getattr(cls, self.partition_key)
        if self.sort_key:
            cls._primary_key = (column, getattr(cls, self.sort_key))
        else:
            cls._primary_key = (column,)
        return cls




from mock import Mock
from tests.helpers.testbase import TestBase
from cerami.datatype import String, Map
from cerami.datatype.mapper import (
    DictMapper)

class TestDictMapper(TestBase):
    def setUp(self):
        self.mocked_map_guesser = Mock()
        self.mocked_parse_guesser = Mock()
        self.dt = Map()
        self.mapper = DictMapper(
            self.dt,
            self.mocked_map_guesser,
            self.mocked_parse_guesser)

    def test__init__(self):
        """it sets guesser"""
        assert self.mapper.map_guesser == self.mocked_map_guesser
        assert self.mapper.parse_guesser == self.mocked_parse_guesser

    def test_resolve(self):
        """it uses guesser to find the correct datatype
        based on the value and calls map for that datatype
        """
        mocked_dt = Mock()
        mocked_dt.mapper.map.return_value = "mocked"
        self.mocked_map_guesser.guess.return_value = mocked_dt
        val = {'test': 1}
        res = self.mapper.resolve(val)
        assert res == {'test': 'mocked'}
        self.mocked_map_guesser.guess.assert_called_with('test', 1)
        mocked_dt.mapper.map.assert_called_with(1)

    def test_parse(self):
        """it uses guesser to find the correct datatype
        based on the value and calls parse for that datatype
        """
        mocked_dt = Mock()
        mocked_dt.mapper.parse.return_value = "mocked"
        self.mocked_parse_guesser.guess.return_value = mocked_dt
        mapped_dict = {'M': {'testkey': {'S': 'testval'}}}
        assert self.mapper.parse(mapped_dict) == {'testkey': 'mocked'}
        self.mocked_parse_guesser.guess.assert_called_with(
            'testkey',
            {'S': 'testval'})
        mocked_dt.mapper.parse.assert_called_with({'S': 'testval'})



from tests.helpers.testbase import TestBase
from cerami.request.search_attribute import (
    UpdateExpressionAttribute,
    UpdateAction)
from cerami.model import Model
from cerami.datatype import String

class FakeModel(Model):
    test = String()
    test2 = String()

class TestUpdateExpressionAttribute(TestBase):
    def setUp(self):
        self.attribute = UpdateExpressionAttribute("Test")

    def test_add_new_action(self):
        """it adds a new key of the action to self.value dict"""

    def test_add(self):
        """it adds the expression to the array"""
        expression = FakeModel.test.eq('test')
        update_action = UpdateAction('SET', expression)
        self.attribute.add(update_action)
        assert self.attribute.value['SET'] == [expression]

    def test_add_appends(self):
        """it appends to the existing value key"""
        expression1 = FakeModel.test.eq('test')
        expression2 = FakeModel.test2.eq('test2')
        update_action1 = UpdateAction('SET', expression1)
        update_action2 = UpdateAction('SET', expression2)
        self.attribute.add(update_action1)
        self.attribute.add(update_action2)
        assert self.attribute.value['SET'] == [expression1, expression2]

    def test_build(self):
        """it returns all group actions comma separated"""
        self.attribute.value = {
            'SET': [
                'fakeExpression1',
                'fakeExpression2'],
            'REMOVE': [
                'fakeExpression3']}
        expected = 'SET fakeExpression1, fakeExpression2 REMOVE fakeExpression3'
        assert self.attribute.build() == expected




from mock import Mock
from tests.helpers.testbase import TestBase
from cerami.response import SearchResponse

class TestSearchResponse(TestBase):
    def setUp(self):
        self.mocked_reconstructor = Mock()
        self.db_response = {
            'Count': 1,
            'ScannedCount': 1,
            'LastEvaluatedKey': 'test',
            'Items': [{'fake': True}]}
        self.resp = SearchResponse(
            self.db_response,
            self.mocked_reconstructor)

    def test__init__(self):
        assert self.resp.count == 1
        assert self.resp.scanned_count == 1
        assert self.resp.last_evaluated_key == 'test'
        assert self.resp._items == [{'fake': True}]

    def test_items(self):
        """it returns an iterable list of reconstructoed items"""
        self.mocked_reconstructor.reconstruct.return_value = 'mocked'
        items = iter(self.resp.items)
        assert next(items) == 'mocked'
        self.mocked_reconstructor.reconstruct.assert_called_with({'fake': True})
        

    def test__init__keyerror(self):
        """it sets items to [] when it is missing"""
        del self.db_response['Items']
        resp = SearchResponse(self.db_response, self.mocked_reconstructor)
        assert resp._items == []



from tests.helpers.testbase import TestBase
from cerami.dynamo_search_interface.search_attribute import AddExpressionAttribute
from cerami.model import Model
from cerami.datatype import Number
from cerami.datatype.expression import UpdateAddExpression

class FakeModel(Model):
    test = Number()

class TestAddExpressionAttribute(TestBase):
    def setUp(self):
        self.attribute = AddExpressionAttribute("Test")

    def test_add(self):
        """it adds the expression to the array"""
        expression = UpdateAddExpression(FakeModel.test, 10)
        self.attribute.add(expression)
        assert self.attribute.value == [expression]

    def test_build(self):
        """returns all expressions joined by ' and '"""
        expression = UpdateAddExpression(FakeModel.test, 10)
        self.attribute.add(expression)
        expected = "ADD {}".format(expression)
        assert self.attribute.build() == expected




from mock import patch, Mock, call
from tests.helpers.testbase import TestBase
from cerami.datatype import String
from cerami.datatype.mapper import (
    BaseDatatypeMapper,
    SetMapperDecorator)

class TestSetMapperDecorator(TestBase):
    def setUp(self):
        self.dt = String()
        self.mapper = BaseDatatypeMapper(self.dt)
        self.mapper.condition_type = 'S'
        self.decorator = SetMapperDecorator(self.mapper)

    def test__init__(self):
        assert self.decorator.mapper == self.mapper

    def test_map(self):
        with patch("cerami.datatype.mapper.SetMapperDecorator.resolve") as resolve:
            resolve.return_value = 'mocked'
            res = self.decorator.map(['test'])
            assert res == {"SS": "mocked"}

    def test_resolve(self):
        """it calls mapper.resolve for each item in value"""
        self.decorator.mapper = Mock()
        self.decorator.mapper.resolve.return_value = "mocked"
        calls = [call(1), call(2)]
        res = self.decorator.resolve([1,2])
        assert res == ['mocked', 'mocked']
        self.decorator.mapper.resolve.assert_has_calls(calls)

    def test_parse(self):
        """assigns the value directly from mapped_dict"""
        mapped_dict = {'SS': ['zac', 'test']}
        assert self.decorator.parse(mapped_dict) == ['zac', 'test']



from .base_datatype_mapper import BaseDatatypeMapper

class ModelMapper(BaseDatatypeMapper):
    def map(self, value):
        mapped = {}
        mapped[self.datatype.condition_type] = value.as_item()
        return mapped

    def resolve(self, value):
        res = {}
        for key, val in value.items():
            column = getattr(self.datatype.model_cls, key)
            res[key] = column.mapper.resolve(val)
        return res

    def parse(self, mapped_dict):
        data = {}
        model_cls = self.datatype.model_cls
        for column in model_cls._columns:
            try:
                val = mapped_dict[self.datatype.condition_type][column.column_name]
            except KeyError:
                continue
            data[column.column_name] = column.mapper.parse(val)
        return model_cls(**data)



from .search_attribute import SearchAttribute

class DictAttribute(SearchAttribute):
    def __init__(self, name, value=None):
        value = value or {}
        super(DictAttribute, self).__init__(name, value)

    def add(self, value_dict):
        self.value.update(value_dict)



from .search_request import SearchRequest
from .search_attribute import DictAttribute
from ..response import SaveResponse

class PutRequest(SearchRequest):
    def execute(self):
        response = self.client.put_item(**self.build())
        return SaveResponse(response, self.reconstructor)

    def item(self, item_dict):
        """add the item_dict to the search_attributes"""
        self.add_attribute(DictAttribute, 'Item', item_dict)
        return self





from .base_datatype_mapper import BaseDatatypeMapper

class ListMapper(BaseDatatypeMapper):
    def __init__(self, datatype, map_guesser, parse_guesser):
        super(ListMapper, self).__init__(datatype)
        self.map_guesser = map_guesser
        self.parse_guesser = parse_guesser

    def resolve(self, value):
        res = []
        for idx, val in enumerate(value):
            guessed_dt = self.map_guesser.guess(idx, val)
            res.append(guessed_dt.mapper.resolve(val))
        return res

    def map(self, value):
        mapped = {}
        res = []
        for idx, val in enumerate(value):
            guessed_dt = self.map_guesser.guess(idx, val)
            res.append(guessed_dt.mapper.map(val))
        mapped[self.datatype.condition_type] = res
        return mapped

    def parse(self, mapped_dict):
        res = []
        items = mapped_dict[self.datatype.condition_type]
        for idx, val in enumerate(items):
            guessed_dt = self.parse_guesser.guess(idx, val)
            res.append(guessed_dt.mapper.parse(val))
        return res



from mock import patch, Mock
from tests.helpers.testbase import TestBase
from cerami.datatype import List, String
from cerami.datatype.expression import ListAppendExpression

class TestListAppendExpression(TestBase):
    def setUp(self):
        self.dt = List(String(), column_name="testcol")
        with patch('cerami.datatype.expression.BaseExpression._generate_variable_name') as gen:
            gen.return_value = "mocked"
            self.expression = ListAppendExpression(self.dt, [1,2,3])

    def test__str__(self):
        """it returns list append structured expression"""
        expected = "#__testcol = list_append(#__testcol, mocked)"
        assert self.expression.__str__() == expected



class DynamoDataAttribute(object):
    def __init__(self, datatype, value=None, initialized=True):
        self.datatype = datatype
        self.value = self.datatype.build(value)
        self.initialized = initialized
        self._changed = False

    def get(self):
        return self.value

    def set(self, newvalue):
        self._changed = True
        self.value = self.datatype.build(newvalue)

    def as_dict(self):
        if self.value is not None:
            return self.datatype.as_dict(self.value)
        return None





from .search_request import SearchRequest
from ..response import SaveResponse
from ..datatype.expression import (
    UpdateRemoveExpression,
    EqualityExpression)
from .search_attribute import (
    DictAttribute,
    UpdateAction,
    UpdateExpressionAttribute)


class UpdateRequest(SearchRequest):
    def execute(self):
        response = self.client.update_item(**self.build())
        return SaveResponse(response, self.reconstructor)

    def key(self, *expressions):
        for expression in expressions:
            key_dict = {}
            key_dict[expression.datatype.column_name] = expression.attribute_map()
            self.add_attribute(DictAttribute, 'Key', key_dict)
        return self

    def set(self, *expressions):
        for expression in expressions:
            self.update_expression('SET', expression)
        return self

    def remove(self, *datatypes):
        for datatype in datatypes:
            expression = UpdateRemoveExpression(datatype)
            self.update_expression('REMOVE', expression)
        return self

    def add(self, datatype, value):
        expression = EqualityExpression('', datatype, value)
        return self.update_expression('ADD', expression)

    def delete(self, datatype, value):
        expression = EqualityExpression('', datatype, value)
        return self.update_expression('DELETE', expression)

    def update_expression(self, action, expression):
        """perform the steps for UpdateExpression"""
        name = {}
        name[expression.expression_attribute_name] = expression.datatype.column_name
        self.add_attribute(UpdateExpressionAttribute,
                           'UpdateExpression',
                           UpdateAction(action, expression))
        self.add_attribute(DictAttribute,
                           'ExpressionAttributeNames',
                           name)
        self.add_attribute(DictAttribute,
                           'ExpressionAttributeValues',
                           expression.value_dict())
        return self



from tests.helpers.testbase import TestBase
from cerami.reconstructor import ModelReconstructor
from cerami.model import Model
from cerami.datatype import String

class TestModelReconstructor(TestBase):
    class TestModel(Model):
        __tablename__ = 'test'
        _id = String()
        test = String()

    def setUp(self):
        self.reconstructor = ModelReconstructor(self.TestModel)

    def test_reconstruct(self):
        """it calls parse for each column"""
        data = {'_id': {'S': '123'}, 'test': {'S': 'cool'}}
        res = self.reconstructor.reconstruct(data)
        assert isinstance(res, self.TestModel)
        assert res._id == '123'
        assert res.test == 'cool'

    def test_reconstruct_missing_key(self):
        """it skips missing columns"""
        data = {'test': {'S': 'cool'}}
        res = self.reconstructor.reconstruct(data)
        assert isinstance(res, self.TestModel)
        assert res._id == None
        assert res.test == 'cool'




from copy import copy
from ..reconstructor import RawReconstructor
from .search_attribute import (
    SearchAttribute,
    DictAttribute,
    QueryExpressionAttribute)

# TODO: break this into classes. Update shouldnt need filter
# and Query should need return_values
class SearchRequest(object):
    def __init__(self, client, tablename="", search_attributes={}, reconstructor=None):
        self.search_attributes = copy(search_attributes)
        self.client = client
        self.reconstructor = reconstructor or RawReconstructor()
        if tablename:
            self.add_attribute(SearchAttribute, 'TableName', tablename)

    def __str__(self):
        return self.build().__str__()

    def add_attribute(self, attr_class, name, value):
        """add a search attribute to a duplicated instance
        @param attr_class - a SearchAttribute class
        @param name - the name of the attribute
        @param value - the value that will be added to the SearchAttribute
        """
        search_attribute = self.search_attributes.get(name, attr_class(name))
        search_attribute.add(value)
        self.search_attributes[name] = search_attribute


    def return_values(self, value):
        self.add_attribute(SearchAttribute,
                           'ReturnValues',
                           value)
        return self

    def filter(self, *expressions):
        """return a new SearchInterface setup with filter attributes
        FilterExpression, ExpressionAttributeNames, ExpressionAttributeValues
        are all required to filter properly
        """
        for expression in expressions:
            names = {}
            names[expression.expression_attribute_name] = expression.datatype.column_name
            self.add_attribute(QueryExpressionAttribute,
                              'FilterExpression',
                              expression)
            self.add_attribute(DictAttribute,
                              'ExpressionAttributeNames',
                              names)
            self.add_attribute(DictAttribute,
                              'ExpressionAttributeValues',
                              expression.value_dict())
        return self

    def build(self):
        """build the dict used by dynamodb"""
        return dict((k, v.build()) for k,v in self.search_attributes.items())




from mock import Mock
from tests.helpers.testbase import TestBase
from cerami.datatype import String, List
from cerami.datatype.mapper import (
    ListMapper)

class TestListMapper(TestBase):
    def setUp(self):
        self.mocked_map_guesser = Mock()
        self.mocked_parse_guesser = Mock()
        self.dt = List()
        self.mapper = ListMapper(
            self.dt,
            self.mocked_map_guesser,
            self.mocked_parse_guesser)

    def test__init__(self):
        """it sets guesser"""
        assert self.mapper.map_guesser == self.mocked_map_guesser
        assert self.mapper.parse_guesser == self.mocked_parse_guesser

    def test_resolve(self):
        """it uses guesser to find the correct datatype
        based on the value and calls map for that datatype
        """
        mocked_dt = Mock()
        mocked_dt.mapper.resolve.return_value = "mocked"
        self.mocked_map_guesser.guess.return_value = mocked_dt
        val = ['test']
        res = self.mapper.resolve(val)
        assert res == ['mocked']
        self.mocked_map_guesser.guess.assert_called_with(0, 'test')
        mocked_dt.mapper.resolve.assert_called_with('test')

    def test_map(self):
        """it uses the guesser to find the correct datatype
        and returns a dict with the mapped value
        """
        mocked_dt = Mock()
        mocked_dt.mapper.map.return_value = "mocked"
        self.mocked_map_guesser.guess.return_value = mocked_dt
        val = ['test']
        res = self.mapper.map(val)
        assert res == {'L': ['mocked']}
        self.mocked_map_guesser.guess.assert_called_with(0, 'test')
        mocked_dt.mapper.map.assert_called_with('test')

    def test_parse(self):
        """it uses guesser to find the correct datatype
        based on the value and calls parse for that datatype
        """
        mocked_dt = Mock()
        mocked_dt.mapper.parse.return_value = "mocked"
        self.mocked_parse_guesser.guess.return_value = mocked_dt
        mapped_dict = {'L': [{'S': 'testval'}]}
        assert self.mapper.parse(mapped_dict) == ['mocked']
        self.mocked_parse_guesser.guess.assert_called_with(
            0,
            {'S': 'testval'})
        mocked_dt.mapper.parse.assert_called_with({'S': 'testval'})



from tests.helpers.testbase import TestBase
from cerami.request.search_attribute import DictAttribute

class TestDictAttribute(TestBase):
    def setUp(self):
        self.attribute = DictAttribute("Test")

    def test_add(self):
        """it updates the value"""
        self.attribute.add({'test': True})
        assert self.attribute.value == {'test': True}
        self.attribute.add({'test': False})
        assert self.attribute.value == {'test': False}

    def test_build(self):
        """returns the value"""
        self.attribute.add({'test': True})
        assert self.attribute.build() == {'test': True}




from .base_expression import BaseExpression

class UpdateRemoveExpression(BaseExpression):
    def __init__(self, datatype):
        super(UpdateRemoveExpression, self).__init__('', datatype, None)

    def value_dict(self):
        """there is no value for this expression so return an empty dict"""
        return {}



from tests.helpers.testbase import TestBase
from cerami.request.search_attribute import QueryExpressionAttribute
from cerami.model import Model
from cerami.datatype import String

class FakeModel(Model):
    test = String()
    test2 = String()

class TestQueryExpressionAttribute(TestBase):
    def setUp(self):
        self.attribute = QueryExpressionAttribute("Test")

    def test_add(self):
        """it adds the expression to the array"""
        expression = FakeModel.test.eq('123')
        self.attribute.add(expression)
        assert self.attribute.value == [expression]

    def test_build(self):
        """returns all expressions joined by ' and '"""
        expr1 = FakeModel.test.eq('123')
        expr2 = FakeModel.test.eq('456')
        self.attribute.add(expr1)
        self.attribute.add(expr2)
        expected = "{} and {}".format(expr1, expr2)
        assert self.attribute.build() == expected



from mock import Mock
from tests.helpers.testbase import TestBase
from cerami.response import GetResponse

class TestGetResponse(TestBase):
    def test__init__(self):
        """it calls reconstruct on item when the key is present"""
        mocked_reconstructor = Mock()
        mocked_reconstructor.reconstruct.return_value = 'mocked'
        db_response = {'Item': {'test': True}}
        resp = GetResponse(db_response, mocked_reconstructor)
        assert resp.item == 'mocked'
        mocked_reconstructor.reconstruct.assert_called_with(db_response['Item'])

    def test__init__keyerror(self):
        """it sets item to None when it is missing from _raw"""
        mocked_reconstructor = Mock()
        db_response = {}
        resp = GetResponse(db_response, mocked_reconstructor)
        assert resp.item == None



from .base_datatype_mapper import BaseDatatypeMapper

class DictMapper(BaseDatatypeMapper):
    def __init__(self, datatype, map_guesser, parse_guesser):
        """initialize the DictMapper
        takes in an additional datatype_guesser
        so it has a way to assigning mapped values
        """
        super(DictMapper, self).__init__(datatype)
        self.map_guesser = map_guesser
        self.parse_guesser = parse_guesser

    def resolve(self, value):
        res = {}
        for key, val in value.items():
            guessed_dt = self.map_guesser.guess(key, val)
            res[key] = guessed_dt.mapper.map(val)
        return res

    def parse(self, mapped_dict):
        res = {}
        data = mapped_dict[self.datatype.condition_type]
        for key, val in data.items():
            guessed_dt = self.parse_guesser.guess(key, val)
            res[key] = guessed_dt.mapper.parse(val)
        return res



class SearchAttribute(object):
    def __init__(self, name, value=None):
        self.name = name
        self.value = value

    def add(self, value):
        self.value = value

    def build(self):
        return self.value



from tests.helpers.testbase import TestBase
from cerami.reconstructor import RawReconstructor

class TestRawReconstructor(TestBase):
    def setUp(self):
        self.reconstructor = RawReconstructor()

    def test_reconstruct(self):
        """it returns the item"""
        item = {'test': True}
        res = self.reconstructor.reconstruct(item)
        assert res == item




from mock import Mock
from tests.helpers.testbase import TestBase
from cerami.response import Response

class TestResponse(TestBase):
    def test__init__(self):
        mocked_reconstructor = Mock()
        db_response = {'test': True}
        resp = Response(db_response, mocked_reconstructor)
        assert resp._raw == db_response
        assert resp.reconstructor == mocked_reconstructor



from .dynamo_data_attribute import DynamoDataAttribute


class DynamoDataListAttribute(DynamoDataAttribute, list):
    def __init__(self, datatype, value=None, initialized=True):
        super(DynamoDataListAttribute, self).__init__(datatype, value, initialized)
        self._appended = []
        self._popped = []

    def __repr__(self):
        return self.value.__repr__()

    def __len__(self):
        return len(self.value)

    def __getitem__(self, key):
        return self.value.__getitem__(key)

    def __setitem__(self, key, value):
        dt = self.datatype.datatype # get Set's datatype
        built = dt.build(attr)
        return self.value.__setitem__(key, built)

    def __deltem__(self, key):
        return self.value.__delitem(key) 

    def __contains__(self, i):
        return self.value.__contains__(i)

    def __length_hint__(self):
        return self.value.__length_hint__()

    def __reversed__(self):
        return self.value.__reversed__()

    def get(self):
        return self

    def append(self, attr):
        return self.insert(len(self), attr)

    def remove(self):
        return self.pop(0)

    def index(self, val):
        return self.value.index(val)

    def count(self, val):
        return self.value.count(val)

    def reverse(self):
        return self.value.reverse()

    def copy(self):
        return self.value.copy()

    def sort(self, resverse=False, key=None):
        return self.value.sort(reverse=reverse, key=key)

    def clear(self):
        self._changed = True
        self._popped.extend(self.copy)
        return self.value.clear()

    def extend(self, values):
        built = self.datatype.build(values)
        self._changed = True
        self._appended.extend(built)
        return self.value.extend(built)

    def insert(self, key, attr):
        dt = self.datatype.datatype # get Set's datatype
        built = dt.build(attr)
        self._changed = True
        self._appended.append(built)
        return self.value.insert(key, built)

    def pop(self, key):
        self._changed = True
        popped = self.value.pop(key)
        self._popped.append(pop)
        return popped



from mock import patch, Mock
from tests.helpers.testbase import TestBase
from cerami.datatype import String
from cerami.datatype.expression import BaseExpression

class TestBaseExpression(TestBase):
    def setUp(self):
        self.dt = String(column_name="testcol")
        with patch('cerami.datatype.expression.BaseExpression._generate_variable_name') as gen:
            gen.return_value = "mocked"
            self.expression = BaseExpression('=', self.dt, 'test')

    def test__init__(self):
        assert self.expression.expression == '='
        assert self.expression.datatype == self.dt
        assert self.expression.value == 'test'
        assert self.expression.expression_attribute_name == '#__testcol'
        assert self.expression.expression_attribute_value_name == 'mocked'

    def test__str__(self):
        """it returns the expression equation"""
        assert self.expression.__str__() == "#__testcol = mocked"

    def test__str__index(self):
        """it includes the index in the string when _index is set on datatype"""
        self.expression.datatype._index = 1
        assert self.expression.__str__() == "#__testcol[1] = mocked"

    def test_attribute_map(self):
        """it calls the datatype mapper with the value"""
        self.expression.datatype = Mock()
        self.expression.attribute_map()
        self.expression.datatype.mapper.map.assert_called_with('test')

    def test_value_dict(self):
        """it returns the dict assigning the condition_type as the
        key to the value"""
        res = self.expression.value_dict()
        assert res == {"mocked": {"S": "test"}}



from .search_request import SearchRequest
from .search_attribute import DictAttribute
from ..response import DeleteResponse

class DeleteRequest(SearchRequest):
    def execute(self):
        response = self.client.delete_item(**self.build())
        return DeleteResponse(response, self.reconstructor)

    def key(self, *expressions):
        for expression in expressions:
            key_dict = {}
            key_dict[expression.datatype.column_name] = expression.attribute_map()
            self.add_attribute(DictAttribute, 'Key', key_dict)
        return self



from .base_expression import BaseExpression

class ListAppendExpression(BaseExpression):
    def __init__(self, datatype, value):
        super(ListAppendExpression, self).__init__( '=', datatype, value)

    def __str__(self):
        attr_name = self.expression_attribute_name
        if hasattr(self.datatype, '_index'):
            attr_name = "{}[{}]".format(attr_name, self.datatype._index)
        return "{attr_name} {expression} list_append({attr_name}, {value_name})".format(
            attr_name=attr_name,
            expression=self.expression,
            value_name=self.expression_attribute_value_name)
